/*
 * DMA on the INCA.
 * Taken from the VxWorks code by Infineon and modified to suit Linux.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>

#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <asm/incaip/dma.h>

/******************************************
*           debug output macros           *
*******************************************/
typedef enum
{
   DMA_ERR,
   DMA_TRACE,
   DMA_FULL
} DMA_DebugLevel_t;

#ifdef DMA_DEBUG_ENABLE
   #define DMA_DEBUG_TX_PRINTOUT

   /*#define  DMA_DEBUG_LEVEL   DMA_ERR*/
   #define  DMA_DEBUG_LEVEL   DMA_FULL

   #define DMA_DEBUG(level,X)          if(level <= DMA_DEBUG_LEVEL) printk X
   #define DMA_LOG(level,S)		printk(S)
   #define DMA_LOG1(level,S,P1)		printk(S,P1)
   #define DMA_LOG2(level,S,P1,P2)	printk(S,P1,P2)
   #define DMA_LOG3(level,S,P1,P2,P3)	printk(S,P1,P2,P3)
#else
   #define DMA_DEBUG(level,X) 
   #define DMA_LOG(level,S)
   #define DMA_LOG1(level,S,P1)
   #define DMA_LOG2(level,S,P1,P2)
   #define DMA_LOG3(level,S,P1,P2,P3)
#endif

/******************************************
*        static variables                 *
*******************************************/
static IncaDmaTxChannel_t  dmaTxChannel[INCA_DMA_MAX_TX_CHANNEL_ID + 1];
static IncaDmaRxChannel_t  dmaRxChannel[INCA_DMA_MAX_RX_CHANNEL_ID + 1];
static int                bDMAinit = 0;

/******************************************
*        local functions and macros       *
*******************************************/
static void incaDMA_RxIsr(int irq, void *dev_id, struct pt_regs *regs);
static void incaDMA_TxIsr(int irq, void *dev_id, struct pt_regs *regs);
static void incaDMA_UpdateHOLD(unsigned long channelId);
static void incaDMA_Receive(unsigned long channelId);

#define DMA_WRITE_REG(reg, value)   *((volatile unsigned long*)reg) = (unsigned long)value;
#define DMA_READ_REG(reg, value)    value = (unsigned long)*((volatile unsigned long*)reg);

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
int incaDMA_Init (void)
{
	int retval;

	if( !bDMAinit )
	{
		/* switch off all DMA channels */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_OFF);

		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, INCA_IP_DMA_DMA_TXCCR1_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, INCA_IP_DMA_DMA_TXCCR2_OFF);

		/* setup TX channel polling time */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXPOLL, INCA_DMA_TX_POLLING_TIME);
		/* setup RX channel polling time */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXPOLL, INCA_DMA_RX_POLLING_TIME);

		/* ERRATA: write reset value into the DMA RX IMR register */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, 0xFFFFFFFF);

		/* install interrupt handlers */
		retval = request_irq(INT_NUM_IM1_IRL27, incaDMA_TxIsr, 0, "dma_tx", NULL);
		if (retval)
			return retval;
		retval = request_irq(INT_NUM_IM1_IRL28, incaDMA_RxIsr, 0, "dma_rx", NULL);
		if (retval)
			return retval;

		bDMAinit = 1;

		DMA_DEBUG(DMA_TRACE,("incaDMA_Init: DMA initialized successfully\n"));
	}
	return 0;
}



/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
/* need these for kfree() in the close routine */
static unsigned long *KMRxfirstDescriptorAddress;
#ifdef INCA_DMA_USE_COPYING
static unsigned char *KMRxBuffers;
#endif

long incaDMA_InitRxChannel ( unsigned long channelId, unsigned long descNum, unsigned long bufSize, 
                              void *device, unsigned long (*getNewBuffer)(void *device),
                              unsigned long (*putRcvdBuffer)(void *device, void *buf, unsigned long bufSize) )
{
   register unsigned long         regValue;
   unsigned long                  *firstDescriptorAddress;
   IncaDmaRxDescriptor_t   *rxDescriptor;
   IncaDmaRxDescriptor_t   *NCrxDescriptor;
   unsigned long                  i;
#ifdef INCA_DMA_USE_COPYING
   unsigned char *RxBuffers;
#endif

   /* check parameters */
   if(channelId > INCA_DMA_MAX_RX_CHANNEL_ID || descNum == 0 || bufSize == 0 ||
      getNewBuffer == NULL || putRcvdBuffer == NULL || device == NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: wrong parameters\n"));
      return -1;
   }

   /* init Rx channel structure */
   dmaRxChannel[channelId].descriptorNumber  = descNum;
   dmaRxChannel[channelId].device            = device;
   dmaRxChannel[channelId].bufferSize        = bufSize;
   dmaRxChannel[channelId].getNewBuffer      = getNewBuffer;
   dmaRxChannel[channelId].putRcvdBuffer     = putRcvdBuffer;


   /* initialize Rx descriptors' list */
#ifdef INCA_DMA_USE_COPYING
   KMRxBuffers = (unsigned char *)kmalloc((descNum * bufSize) + 16, GFP_KERNEL);
   if (KMRxBuffers == NULL)
   	return -1;
   RxBuffers = KMRxBuffers;
   if ((unsigned long)RxBuffers & 0xf)
   	RxBuffers = (unsigned char *)(((unsigned long)RxBuffers + 16) & 0xfffffff0);

   if((KMRxfirstDescriptorAddress =  (unsigned long*)kmalloc(descNum * 
                                 (sizeof(IncaDmaRxDescriptor_t))+16, GFP_KERNEL)) == NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   }

#else
   if((KMRxfirstDescriptorAddress = (unsigned long*)kmalloc(descNum * sizeof(IncaDmaRxDescriptor_t)+16, GFP_KERNEL))==NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   }
#endif

   firstDescriptorAddress = KMRxfirstDescriptorAddress;
   /* Make 16 bytes alingment of descriptor addresses */
   if((unsigned long)firstDescriptorAddress & 0xF)
   {
      firstDescriptorAddress += 4;
      firstDescriptorAddress = (unsigned long*)((unsigned long)firstDescriptorAddress & 0xFFFFFFF0);
   }

#ifndef CONFIG_MIPS_UNCACHED
   firstDescriptorAddress = (unsigned long*)ioremap_nocache(virt_to_phys(firstDescriptorAddress), sizeof(IncaDmaRxDescriptor_t));
#endif
   dmaRxChannel[channelId].firstDescriptorAddr = firstDescriptorAddress;

   /* clear memory area */
#ifdef INCA_DMA_USE_COPYING
   memset(firstDescriptorAddress, 0, descNum * (sizeof(IncaDmaRxDescriptor_t)));
   memset(RxBuffers, 0, descNum * bufSize);
#else
   memset(firstDescriptorAddress, 0, descNum * sizeof(IncaDmaRxDescriptor_t));
#endif

   for(i = 0, rxDescriptor = (IncaDmaRxDescriptor_t*)firstDescriptorAddress;
       i < descNum;
       i++, rxDescriptor++)
   {

      NCrxDescriptor = (IncaDmaRxDescriptor_t *)KSEG1ADDR(rxDescriptor);
      /* set maximum size of receive buffer */
      NCrxDescriptor->params.field.NFB =  dmaRxChannel[channelId].bufferSize /* TODO: use correct value */;

      /* set the offset of the receive buffer */
      NCrxDescriptor->params.field.offset = 0; /* offset mechanism is not used */

      /* check if it is the last descriptor */
      if( i == (descNum - 1) )
      {
         /* let the last descriptor point to the first one */
         NCrxDescriptor->nextRxDescPtr = (unsigned long)firstDescriptorAddress;
      }
      else
      {
         /* set the address of the next descriptor */
#ifndef CONFIG_MIPS_UNCACHED
         NCrxDescriptor->nextRxDescPtr =
		(unsigned long)ioremap_nocache(virt_to_phys(rxDescriptor + 1), sizeof(IncaDmaRxDescriptor_t));
#else
         NCrxDescriptor->nextRxDescPtr = (unsigned long)(rxDescriptor + 1);
#endif
      }

#ifdef INCA_DMA_USE_COPYING
#ifndef CONFIG_MIPS_UNCACHED
      NCrxDescriptor->RxDataPtr = (unsigned long)ioremap_nocache(virt_to_phys(RxBuffers + i * bufSize), bufSize);
#else
      NCrxDescriptor->RxDataPtr = (unsigned long)RxBuffers + i * bufSize;
#endif
#else
      /* Allocate memory for data buffers */
      NCrxDescriptor->RxDataPtr = getNewBuffer(device);
#endif
      
      if(NCrxDescriptor->RxDataPtr == 0)
      {
         DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: cannot allocate buffer for descriptor\n"));
         return -1;
      }

      NCrxDescriptor->params.field.IEop = 1;
   }

   /* set HOLD bit in last decriptor.
      and store last descriptor in the channel object */
#ifndef CONFIG_MIPS_UNCACHED
   dmaRxChannel[channelId].holdDescriptor =
		ioremap_nocache(virt_to_phys(rxDescriptor - 1), sizeof(IncaDmaRxDescriptor_t));
#else
   dmaRxChannel[channelId].holdDescriptor = rxDescriptor - 1;
#endif
   dmaRxChannel[channelId].holdDescriptor->params.field.HOLD = 1;

   dmaRxChannel[channelId].firstDescriptor = (IncaDmaRxDescriptor_t*)firstDescriptorAddress;
   dmaRxChannel[channelId].currentDescriptor = (IncaDmaRxDescriptor_t*)firstDescriptorAddress;

   /* Writing to the MASK register: unmask the HLD, CMDCPT 
                                    mask the EOP, CPT and RDERR*/
   DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, regValue);
   regValue &= ~( INCA_IP_DMA_RXIMR_HLD(channelId)    | 
                  INCA_IP_DMA_RXIMR_CMDCPT(channelId) |
                  INCA_IP_DMA_RXIMR_CPT(channelId)    |
                  INCA_IP_DMA_RXIMR_EOP(channelId)    |
                  INCA_IP_DMA_RXIMR_RDERR(channelId)  );
   DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, regValue);

   switch(channelId)
   {
      case 0:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXFRDA0, firstDescriptorAddress);

         /* Writing to the COMMAND REG */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_INIT);
         break;

      case 1:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXFRDA1, firstDescriptorAddress);

         /* Writing to the COMMAND REG */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_INIT);
         break;
      
      default: 
         break;
   }

   /* enable interrupts */
   DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMARIR);
   DMA_INTERRUPTS_ENABLE(INCA_IRQ_LINE_DMARIR);

   DMA_DEBUG(DMA_TRACE,("incaDMA_InitRxChannel: Rx channel %d initialized\n",channelId));
   DMA_DEBUG(DMA_TRACE,("                       Number of Descriptors: %d\n",descNum));
   return 0;
}



/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
/* need this for kfree() in the close routine */
static unsigned long *KMTxfirstDescriptorAddress;

long incaDMA_InitTxChannel ( unsigned long channelId, unsigned long descNum, unsigned long priority, 
                              void *device, unsigned long (*txComplete)(void* device) )
{
   register unsigned long         regValue;
   unsigned long                  i;  
   unsigned long                  *firstDescriptorAddress;
   IncaDmaTxDescriptor_t   *txDescriptor;
   IncaDmaTxDescriptor_t   *NCtxDescriptor;

   /* check parameters */
   if(   channelId > INCA_DMA_MAX_TX_CHANNEL_ID || 
         priority > INCA_DMA_MAX_TX_PRIORITY    ||
         device == NULL                         ||
         txComplete == NULL                     ||
         descNum == 0   )
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitTxChannel: wrong parameters\n"));
      return -1;
   }

   /* init Tx channel structure */
   dmaTxChannel[channelId].descriptorNumber  = descNum;
   dmaTxChannel[channelId].device            = device;
   dmaTxChannel[channelId].txComplete        = txComplete;

   /* initialize Tx descriptors' list */
#ifdef INCA_DMA_USE_COPYING
   if((KMTxfirstDescriptorAddress =  (unsigned long*)kmalloc(descNum * 
                                 (sizeof(IncaDmaTxDescriptor_t) + 1520)+16, GFP_KERNEL))==NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitTxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   }
#else
   if((KMTxfirstDescriptorAddress = (unsigned long*)kmalloc(descNum * sizeof(IncaDmaTxDescriptor_t)+16, GFP_KERNEL))==NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitTxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   }
#endif
   firstDescriptorAddress = KMTxfirstDescriptorAddress;
   /* Make 16 bytes alingment of descriptor addresses */
   if((unsigned long)firstDescriptorAddress & 0xF)
   {
      firstDescriptorAddress += 4;
      firstDescriptorAddress = (unsigned long*)((unsigned long)firstDescriptorAddress & 0xFFFFFFF0);
   }
#ifndef CONFIG_MIPS_UNCACHED
   firstDescriptorAddress = (unsigned long*)ioremap_nocache(virt_to_phys(firstDescriptorAddress), sizeof(IncaDmaTxDescriptor_t));
#endif
   dmaTxChannel[channelId].firstDescriptorAddr = firstDescriptorAddress;

   /* clear memory area */
   memset(firstDescriptorAddress, 0, descNum * sizeof(IncaDmaTxDescriptor_t));

   /* Build a linked list of  Tx descriptors */
   for(i = 0, txDescriptor = (IncaDmaTxDescriptor_t*)firstDescriptorAddress;
       i < descNum;
       i++, txDescriptor++)
   {
      NCtxDescriptor = (IncaDmaTxDescriptor_t *)KSEG1ADDR(txDescriptor);

      NCtxDescriptor->params.dword       = 0;
      NCtxDescriptor->params.field.HOLD  = 1;
      NCtxDescriptor->C                  = 1;

#ifdef INCA_DMA_USE_COPYING
#ifndef CONFIG_MIPS_UNCACHED
      NCtxDescriptor->TxDataPtr = (unsigned long)ioremap_nocache(virt_to_phys((void *)((unsigned long)firstDescriptorAddress + descNum * sizeof(IncaDmaTxDescriptor_t) + 1520*i)), 1520);
#else
      NCtxDescriptor->TxDataPtr = (unsigned long)firstDescriptorAddress + 
                                 descNum * sizeof(IncaDmaTxDescriptor_t) + 1520*i;
#endif
#endif      

      /* check if it is the last descriptor */
      if( i == (descNum - 1) )
      {
         /* let the last descriptor points to the first one */
         NCtxDescriptor->nextTxDescPtr = (unsigned long)firstDescriptorAddress;
      }
      else
      {
         /* set the address of the next descriptor */
#ifndef CONFIG_MIPS_UNCACHED
         NCtxDescriptor->nextTxDescPtr =
		(unsigned long)ioremap_nocache(virt_to_phys(txDescriptor + 1), sizeof(IncaDmaTxDescriptor_t));
#else
         NCtxDescriptor->nextTxDescPtr = (unsigned long)(txDescriptor + 1);
#endif
      }
   }

   /* Store the current hold descriptor in the channel object */
#ifndef CONFIG_MIPS_UNCACHED
   dmaTxChannel[channelId].holdDescriptor =
		ioremap_nocache(virt_to_phys(txDescriptor - 1), sizeof(IncaDmaTxDescriptor_t));
#else
   dmaTxChannel[channelId].holdDescriptor = txDescriptor - 1;
#endif

   /* Writing a priority of the channel (0-3)  */
   DMA_READ_REG (INCA_IP_DMA_DMA_TXWGT, regValue);
   regValue |= INCA_IP_DMA_TXWGT_TXPRI(channelId,priority);
   DMA_WRITE_REG (INCA_IP_DMA_DMA_TXWGT, regValue);

   /* Writing to the MASK register: unmask the HLD, CMDCPT, CPT
                                    mask the EOP, RDERR */
   DMA_READ_REG (INCA_IP_DMA_DMA_TXIMR, regValue);
   regValue &= ~( INCA_IP_DMA_TXIMR_HLD(channelId)    |   
                  INCA_IP_DMA_TXIMR_CMDCPT(channelId) |
                  INCA_IP_DMA_TXIMR_CPT(channelId)    |
                  INCA_IP_DMA_TXIMR_EOP(channelId)    |
                  INCA_IP_DMA_TXISR_RDERR(channelId)  );
   DMA_WRITE_REG (INCA_IP_DMA_DMA_TXIMR, regValue);

   /* enable interrupts */
   DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMATIR);
   DMA_INTERRUPTS_ENABLE(INCA_IRQ_LINE_DMATIR);

   switch(channelId)
   {
      case 0:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA0, firstDescriptorAddress);
         break;

      case 1:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA1, firstDescriptorAddress);
         break;

      case 2:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA2, firstDescriptorAddress);
         break;

      default: 
         break;
   }

   DMA_DEBUG(DMA_TRACE,("incaDMA_InitTxChannel: Tx channel %d initialized\n",channelId));
   DMA_DEBUG(DMA_TRACE,("                       Number of Descriptors: %d\n",descNum));
   DMA_DEBUG(DMA_TRACE,("                       Priority: %d\n",priority));

   return 0;
}


/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
void incaDMA_RxIsr(int irq, void *dev_id, struct pt_regs *regs)
{
   register unsigned long regValue;
   /* Acknowledge the interrupt */
   /*DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMARIR);*/

   /* Check DMA_RXISR register */
   DMA_READ_REG(INCA_IP_DMA_DMA_RXISR,regValue);

   if(regValue == 0)
      return;

   DMA_LOG1(DMA_TRACE,"incaDMA_RxIsr: (0x%08X)\n",regValue);

   /* Acknowledge interrupts */
/*   DMA_WRITE_REG(INCA_IP_DMA_RXISR,regValue);*/
   DMA_WRITE_REG(INCA_IP_DMA_DMA_RXISR,0xFFFFFFFF);

   /* What kind of interrupt we've got */
   if(regValue & INCA_IP_DMA_RXISR_HLD_0)
   {
      DMA_LOG(DMA_TRACE,"  HOLD (0)\n");
      incaDMA_UpdateHOLD(0);
   }
   if(regValue & INCA_IP_DMA_RXISR_HLD_1)
   {
      DMA_LOG(DMA_TRACE,"  HOLD (1)\n");
      incaDMA_UpdateHOLD(1);
   }


   if(regValue & INCA_IP_DMA_RXISR_EOP_0)
   {
      /* End Of Packet */
      /* We have to pass data to the upper layer */
      DMA_LOG(DMA_TRACE,"  EOP (0)\n");
      incaDMA_Receive(0);
   }
   if(regValue & INCA_IP_DMA_RXISR_EOP_1)
   {
      /* End Of Packet */
      /* We have to pass data to the upper layer */
      DMA_LOG(DMA_TRACE,"  EOP (1)\n");
      incaDMA_Receive(1);
   }

#ifdef DMA_DEBUG_ENABLE
   if(regValue & INCA_IP_DMA_RXISR_CPT_0)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_RXISR_CPT_1)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (1)\n");
   }


   if(regValue & INCA_IP_DMA_RXISR_CMDCPT_0)
   {
      /* Command Complete */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CMDCPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_RXISR_CMDCPT_1)
   {
      /* Command Complete */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CMDCPT (1)\n");
   }


   if(regValue & INCA_IP_DMA_RXISR_RDERR_0)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (0)\n");
   }
   if(regValue & INCA_IP_DMA_RXISR_RDERR_1)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (1)\n");
   }
#endif /* DMA_DEBUG_ENABLE */
}



/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_TxIsr(int irq, void *dev_id, struct pt_regs *regs)
{
   register unsigned long regValue;

   /* Acknowledge the interrupt */
   /*DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMATIR);*/
   /* Check DMA_TXISR register */
   DMA_READ_REG(INCA_IP_DMA_DMA_TXISR,regValue);

   DMA_LOG1(DMA_TRACE,"incaDMA_TxIsr: (0x%08X)\n",regValue);

   /* acknowledge interrupt */
   /*DMA_WRITE_REG(INCA_IP_DMA_DMA_TXISR,regValue);*/
   DMA_WRITE_REG(INCA_IP_DMA_DMA_TXISR,0xFFFFFFFF);

   if(regValue == 0)
      return;

   /* What kind of interrupt we've got */
   if(regValue & INCA_IP_DMA_TXISR_EOP_0)
   {
      /* End Of Packet */
      /* Packet is transmitted so transmit next one */
      DMA_LOG(DMA_TRACE,"  EOP (0)\n");
      if(dmaTxChannel[0].txComplete)
      {
         DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[0].txComplete);
         dmaTxChannel[0].txComplete(dmaTxChannel[0].device);
      }
      dmaTxChannel[0].txAckCounter ++;
   }

   if(regValue & INCA_IP_DMA_TXISR_EOP_1)
   {
      /* End Of Packet */
      /* Packet is transmitted so transmit next one */
      DMA_LOG(DMA_TRACE,"  EOP (1)\n");
      if(dmaTxChannel[1].txComplete)
      {
         DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[1].txComplete);
         dmaTxChannel[1].txComplete(dmaTxChannel[1].device);
      }
      dmaTxChannel[1].txAckCounter ++;
   }
   if(regValue & INCA_IP_DMA_TXISR_EOP_2)
   {
      /* End Of Packet */
      /* Packet is transmitted so transmit next one */
      DMA_LOG(DMA_TRACE,"  EOP (2)\n");
      if(dmaTxChannel[2].txComplete)
      {
         DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[2].txComplete);
         dmaTxChannel[2].txComplete(dmaTxChannel[2].device);
      }
      dmaTxChannel[2].txAckCounter ++;
   }


#ifdef DMA_DEBUG_ENABLE
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_0)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_1)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_2)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (2)\n");
   }

   if(regValue & INCA_IP_DMA_TXISR_CPT_0)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CPT_1)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CPT_2)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (2)\n");
   }


   if(regValue & INCA_IP_DMA_TXISR_HLD_0)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_HLD_1)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_HLD_2)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (2)\n");
   }

 
   if(regValue & INCA_IP_DMA_TXISR_RDERR_0)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_RDERR_1)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_RDERR_2)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (2)\n");
   }
#endif /* DMA_DEBUG_ENABLE */
}


/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_UpdateHOLD(unsigned long channelId)
{
   register unsigned long         regValue;
   IncaDmaRxDescriptor_t   *newHoldDesc;

   /* find out the previous descriptor
      If the HOLD descriptor is the first one, set it to the last one */
   if(dmaRxChannel[channelId].holdDescriptor == dmaRxChannel[channelId].firstDescriptor)
   {
      newHoldDesc = dmaRxChannel[channelId].firstDescriptor + 
                                             (dmaRxChannel[channelId].descriptorNumber-1);
   }
   else
   {
      /* if not, it is the previous one, i.e. 4 words before*/
      newHoldDesc = dmaRxChannel[channelId].holdDescriptor--;
   }


   if(newHoldDesc->status.field.C == 0)
   {
      /* set the HOLD bit in the new HOLD descriptor */
      newHoldDesc->params.field.HOLD = 1;

      /* update a HOLD descriptor */
      dmaRxChannel[channelId].holdDescriptor->params.field.HOLD = 0;
      dmaRxChannel[channelId].holdDescriptor = newHoldDesc;
   }
   else
   {
      DMA_LOG(DMA_ERR,"incaDMA_UpdateHOLD: can't do it\n");
   }

   /* and give a HOLD RESET command */
   if( channelId == 0 )
   {
      DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
      regValue |= INCA_IP_DMA_DMA_RXCCR0_HR;
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
   }
   else
   {
      DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
      regValue |= INCA_IP_DMA_DMA_RXCCR1_HR;
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
   }
}


/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_Receive(unsigned long channelId)
{
   IncaDmaRxDescriptor_t   *pCurrentDesc;
   unsigned long                  newBuffer, result, counter;

   pCurrentDesc = dmaRxChannel[channelId].currentDescriptor;
   if (pCurrentDesc == NULL)
   {
      return;
   }

   if(dmaRxChannel[channelId].getNewBuffer == NULL || 
      dmaRxChannel[channelId].putRcvdBuffer == NULL )
   {
      return;
   }
   
   counter = 0;
   while( pCurrentDesc->status.dword & (INCA_DMA_RX_EOP | INCA_DMA_RX_C) )
   {
#ifdef INCA_DMA_USE_COPYING
      newBuffer = pCurrentDesc->RxDataPtr;   
#else
      newBuffer = dmaRxChannel[channelId].getNewBuffer(dmaRxChannel[channelId].device);
#endif
      if(newBuffer == 0)
      {
         return;
      }

      /* reset Hold RX descriptor */
      pCurrentDesc->params.field.HOLD = 1;

      pCurrentDesc->status.dword &= ~(INCA_DMA_RX_EOP | INCA_DMA_RX_SOP | INCA_DMA_RX_C);

      dmaRxChannel[channelId].holdDescriptor->params.field.HOLD = 0;
      dmaRxChannel[channelId].holdDescriptor = pCurrentDesc;

      DMA_LOG2(DMA_ERR,"DMA_RX: 0x%08X %d\n",pCurrentDesc->RxDataPtr,pCurrentDesc->status.field.NBT);


#undef DMA_DEBUG_RX_PRINTOUT
#ifdef DMA_DEBUG_RX_PRINTOUT
   for (result = 0; result < pCurrentDesc->status.field.NBT-4; result++)
   {
      printk("R0x%02X ",((unsigned char*)pCurrentDesc->RxDataPtr)[result]);
      if(((result+1)%16) == 0)printk("\n");
   }
   printk("\n");
#endif

      result = dmaRxChannel[channelId].putRcvdBuffer(
                                 dmaRxChannel[channelId].device,
                                 (void*)pCurrentDesc->RxDataPtr,
                                 pCurrentDesc->status.field.NBT-4);
      if(result == -1)
      {
         DMA_LOG(DMA_ERR,"incaDMA_Receive: up layer error \n");
      }                                            

      pCurrentDesc->RxDataPtr = newBuffer;
      pCurrentDesc = (IncaDmaRxDescriptor_t*)pCurrentDesc->nextRxDescPtr;
      
      counter++;
   }

   dmaRxChannel[channelId].currentDescriptor = (IncaDmaRxDescriptor_t*)pCurrentDesc;
   dmaRxChannel[channelId].rxCounter ++;
   
   if( counter > dmaRxChannel[channelId].maxRxInOneIrq )
   {
      dmaRxChannel[channelId].maxRxInOneIrq = counter;
   }
}



/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
unsigned long incaDMA_Transmit(unsigned long channelId, void *buffer, unsigned long bufLen, int lastBuffer)
{
   register unsigned long   regValue, command;
   int              firstBuffer = 0;

   if(buffer == NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: wrong parameters\n"));
      return -1;
   }

    /* get the current hold descriptor */
   if(dmaTxChannel[channelId].currentDescriptor == NULL)
   {
      /* this is a first frame of the packet */
      /* get the next free descriptor */
      dmaTxChannel[channelId].currentDescriptor = 
         (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].holdDescriptor->nextTxDescPtr;
      firstBuffer = 1;

      dmaTxChannel[channelId].curPacketSize = 0;

#ifdef INCA_DMA_USE_COPYING
      dmaTxChannel[channelId].currentBufferPtr = (unsigned char*)dmaTxChannel[channelId].currentDescriptor->TxDataPtr;
   }
#else
   }
   else
   {
      /* get the next free descriptor */
      dmaTxChannel[channelId].currentDescriptor = 
         (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].currentDescriptor->nextTxDescPtr;
   }

   /* Check if the complete bit is already set. If not, the descriptor is
      still under control of the DMA and cannot be used. This means that 
      there is no free descriptor and we have reject the transmit attempt. */
   if( dmaTxChannel[channelId].currentDescriptor->C == 0 ) 
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: there is no free descriptor\n"));
      dmaTxChannel[channelId].currentDescriptor = NULL;
      return -1;
   }

   /* Set the TX data pointer */
   dmaTxChannel[channelId].currentDescriptor->TxDataPtr = (unsigned long)(buffer);
   /* Set the buffer size */
   dmaTxChannel[channelId].currentDescriptor->params.field.NBA = bufLen;
#endif

#ifdef INCA_DMA_USE_COPYING
   /***********************************************************************/
   /* copy data */
   if(dmaTxChannel[channelId].currentBufferPtr == NULL)
   {
      return -1;
   }

   memcpy(dmaTxChannel[channelId].currentBufferPtr,buffer,bufLen);
   dmaTxChannel[channelId].currentBufferPtr += bufLen; 
   dmaTxChannel[channelId].curPacketSize += bufLen;

   /* Set the buffer size */
   dmaTxChannel[channelId].currentDescriptor->params.field.NBA = dmaTxChannel[channelId].curPacketSize;
   /***********************************************************************/
#endif 
      
#undef DMA_DEBUG_TX_PRINTOUT
#ifdef DMA_DEBUG_TX_PRINTOUT
   for (command = 0; command < bufLen; command++)
   {
      printk("T0x%02X ",((unsigned char*)dmaTxChannel[channelId].currentDescriptor->TxDataPtr)[command]);
      if(((command+1)%16) == 0)printk("\n");
   }
   printk("\n");
#endif

   /* delete Complete bit so that it can be set again by DMA */
   dmaTxChannel[channelId].currentDescriptor->C = 0;

   if(firstBuffer == 1)
   {
      dmaTxChannel[channelId].currentDescriptor->params.field.Sop    = 1;
/*      DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit(%d): Tx first frame of the packet\n", channelId));*/
   }
   else
   {
      dmaTxChannel[channelId].currentDescriptor->params.field.Sop    = 0;
   }
   
   if(lastBuffer == 1)
   {
/*      DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit(%d): Tx last frame of the packet\n", channelId));*/
      /* If it is the last buffer of a frame 
         then set FE , HOLD , THI bits
         so that the DT stops after transmit */
      dmaTxChannel[channelId].currentDescriptor->params.dword |= 
                  ( INCA_DMA_TX_EOP | INCA_DMA_TX_IEOP | INCA_DMA_TX_HOLD );

      /* Reset the HOLD bit */
      dmaTxChannel[channelId].holdDescriptor->params.field.HOLD = 0;

      /* update the TX hold descriptor */
      dmaTxChannel[channelId].holdDescriptor = dmaTxChannel[channelId].currentDescriptor;
      dmaTxChannel[channelId].currentDescriptor = NULL;

      /* set the command to force a re-read of the hold descriptor by the DMA */
      /* and give a HOLD RESET command */
      if(dmaTxChannel[channelId].bInit == 0)
      {
         command = INCA_IP_DMA_DMA_TXCCR0_INIT;
         dmaTxChannel[channelId].bInit = 1;
      }
      else
      {
         command = INCA_IP_DMA_DMA_TXCCR0_HR;
      }
      
      dmaTxChannel[channelId].txCounter ++;
      if(dmaTxChannel[channelId].curPacketSize > dmaTxChannel[channelId].maxPacketSize)
      {
         dmaTxChannel[channelId].maxPacketSize = dmaTxChannel[channelId].curPacketSize;
      }
   
      switch(channelId)
      {
         case 0:
            DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR0, regValue);
            regValue |= command;
            DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, regValue);
            break;

         case 1:
            DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR1, regValue);
            regValue |= command;
            DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, regValue);
            break;

         case 2:
            DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR2, regValue);
            regValue |= command;
            DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, regValue);
            break;

         default:
            break;
      }
   }
   else
   {
      dmaTxChannel[channelId].currentDescriptor->params.dword &= 
                  ~( INCA_DMA_TX_EOP | INCA_DMA_TX_IEOP | 
                     INCA_DMA_TX_HOLD | INCA_DMA_TX_ICPT );
   }

#ifdef DMA_DEBUG_ENABLE 
/*   printk("DMA_TX end:\n");
   if(dmaTxChannel[channelId].currentDescriptor)
      incaDMA_PrintTxInfo( dmaTxChannel[channelId].currentDescriptor );
   else
      incaDMA_PrintTxInfo( dmaTxChannel[channelId].holdDescriptor );*/
#endif

/*   DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit: Id:%d, size:%d\n", channelId, bufLen));*/
   dmaTxChannel[channelId].useCounter ++;

   return 0;
}


/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
long incaDMA_CloseTxChannel ( unsigned long channelId )
{
   /* close channel */
   switch(channelId)
   {
      case 0:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, 0x00000000);
         break;

      case 1:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, 0x00000000);
         break;

      case 2:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, 0x00000000);
         break;

      default:
         return -1;
   }

   /* release all resources */
   if(dmaTxChannel[channelId].firstDescriptorAddr)
   {
      kfree(KMTxfirstDescriptorAddress);

      /* uninit Rx channel structure */
      memset(&dmaTxChannel[channelId],0,sizeof(IncaDmaTxChannel_t));
   }
   return 0;
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
long incaDMA_CloseRxChannel ( unsigned long channelId )
{
   /* close channel */
   switch(channelId)
   {
      case 0:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, 0x00000000);
         break;

      case 1:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, 0x00000000);
         break;

      default:
         return -1;
   }

   /* release all resources */
   if (dmaRxChannel[channelId].firstDescriptorAddr)
   {
      kfree(KMRxfirstDescriptorAddress);
#ifdef INCA_DMA_USE_COPYING
      kfree(KMRxBuffers);
#endif

      /* uninit Rx channel structure */
      memset(&dmaRxChannel[channelId],0,sizeof(IncaDmaRxChannel_t));
   }

   return 0;
}

#ifdef DMA_DEBUG_ENABLE

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
void incaDMA_PrintAllInfoRx(unsigned long channelId)
{
   IncaDmaRxDescriptor_t   *rxDescriptor;
   unsigned long                  i;

   printk("------------------------  RX DESCRIPTORS  --------------------------\n");
   for(i = 0, rxDescriptor = dmaRxChannel[channelId].firstDescriptor;
       i < dmaRxChannel[channelId].descriptorNumber;
       i++, rxDescriptor = (IncaDmaRxDescriptor_t*)rxDescriptor->nextRxDescPtr)
   {
      incaDMA_PrintRxInfo( rxDescriptor );
   }
   printk("--------------------------------------------------------------------\n");
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
void incaDMA_PrintAllInfoTx(unsigned long channelId)
{
   IncaDmaTxDescriptor_t   *txDescriptor;                     
   unsigned long                  i;

   printk("------------------------  TX DESCRIPTORS  --------------------------\n");
   for(i = 0, txDescriptor = (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].holdDescriptor->nextTxDescPtr;
       i < dmaTxChannel[channelId].descriptorNumber;
       i++, txDescriptor = (IncaDmaTxDescriptor_t*)txDescriptor->nextTxDescPtr)
   {
      incaDMA_PrintTxInfo( txDescriptor );
   }
   printk("--------------------------------------------------------------------\n");
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
void incaDMA_PrintTxInfo( IncaDmaTxDescriptor_t *desc )
{
   printk("0x%08X> HOLD(%d), Eop(%d), Sop(%d), ICpt(%d), IEop(%d), NBA(%d)\n",
         desc,
         desc->params.field.HOLD,
         desc->params.field.Eop,
         desc->params.field.Sop,
         desc->params.field.ICpt,
         desc->params.field.IEop,
         desc->params.field.NBA);

   printk("            next(0x%08X),  data(0x%08X), C(%d)\n",
         desc->nextTxDescPtr,
         desc->TxDataPtr,
         desc->C);
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
void incaDMA_PrintRxInfo( IncaDmaRxDescriptor_t *desc )
{
   printk("0x%08X> HOLD(%d), ICpt(%d), IEop(%d), Sop(%d), Eop(%d), NFB(%d), NBT(%d)\n",
         desc,
         desc->params.field.HOLD,
         desc->params.field.ICpt,
         desc->params.field.IEop,

         desc->status.field.Sop,
         desc->status.field.Eop,
         desc->params.field.NFB,
         desc->status.field.NBT);

   printk("            next(0x%08X),  data(0x%08X), C(%d)\n",
         desc->nextRxDescPtr,
         desc->RxDataPtr,
         desc->status.field.C);
}

#endif /*DMA_DEBUG_ENABLE*/
