/*
 * Handle mapping of the flash memory access routines
 * on INCA-IP based devices.
 *
 * Copyright(C) 2003 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This code is GPLed
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/cfi.h>

#undef	DEBUG

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif


#define FLASH_BANK_MAX 2

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

/* static struct mtd_info *mymtd; */
static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
static unsigned long num_banks;
static unsigned long start_scan_addr;
static int probing = 0;

__u8 incaip_read8(struct map_info *map, unsigned long ofs)
{
	return *((__u8 *)(map->map_priv_1 + ofs));
}

__u16 incaip_read16(struct map_info *map, unsigned long ofs)
{
 /* Modify the address offset to account the EBU address swap.
  * Needed only for probing.
  */
	if (probing)
		ofs ^= 2;
	return *((__u16 *)(map->map_priv_1 + ofs));
}

__u32 incaip_read32(struct map_info *map, unsigned long ofs)
{
	return *((__u32 *)(map->map_priv_1 + ofs));
}

void incaip_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

void incaip_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	*((__u8 *)(map->map_priv_1 + adr)) = d;
}

void incaip_write16(struct map_info *map, __u16 d, unsigned long adr)
{
 /* Modify the address offset to account the EBU address swap.
  * Needed only for probing.
  */
	if (probing)
		adr ^= 2;
	*((__u16 *)( map->map_priv_1 + adr)) = d;
}

void incaip_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	*((__u32 *)(map->map_priv_1 + adr)) = d;
}

void incaip_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

struct map_info incaip_map = {
	name:		"INCA-IP",
	/* size:	WINDOW_SIZE, */
	buswidth:	2,
	read8:		incaip_read8,
	read16:		incaip_read16,
	read32:		incaip_read32,
	copy_from:	incaip_copy_from,
	write8:		incaip_write8,
	write16:	incaip_write16,
	write32:	incaip_write32,
	copy_to:	incaip_copy_to
};

/*
 * Here are partition information for all known INCA-IP series devices.
 * See include/linux/mtd/partitions.h for definition of the mtd_partition
 * structure.
 */

#ifdef CONFIG_MTD_PARTITIONS

/* partition definition for first flash bank
 * also ref. to "drivers/char/flash_config.c"
 */
static struct mtd_partition incaip_partitions[] = {
	{
	  name:		"U-Boot",	/* U-Boot firmware */
	  offset:	0x00000000,
	  size:		0x00030000,	/* 192KB	   */
/*	  mask_flags:	MTD_WRITEABLE,	   force read-only */
	},
	{
	  name:		"Environment",	/* default kernel image */
	  offset:	0x00030000,
	  size:		0x00010000,
/*	  mask_flags:	MTD_WRITEABLE,	   force read-only */
	},
	{
	  name:		"Linux",	/* default kernel image */
	  offset:	0x00040000,
	  size:		0x000C0000,
/*	  mask_flags:	MTD_WRITEABLE,     force read-only */
	},
	{
	  name:		"rootfs",	/* default root filesystem */
	  offset:	0x00100000,
	  size:		0x00300000,
/*	  mask_flags:	MTD_WRITEABLE,     force read-only */
	},
	{
	  name:		"Linux 2",	/* alternate kernel image */
	  offset:	0x00400000,
	  size:		0x000C0000,
/*	  mask_flags:	MTD_WRITEABLE,     force read-only */
	},
	{
	  name:		"rootfs 2",	/* alternate root filesystem */
	  offset:	0x00500000,
	  size:		0x00300000,
/*	  mask_flags:	MTD_WRITEABLE,     force read-only */
	},
};
/* partition definition for second flash bank */
static struct mtd_partition incaip_fs_partitions[] = {
	{
	  name:		"User Data",
	  offset:	0x00000000,
	  size:		MTDPART_SIZ_FULL,
	}
};
#endif	/* CONFIG_MTD_PARTITIONS */

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

unsigned long flash_start, flash_size;

int __init init_incaip_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long mtd_size = 0;

	/* request maximum flash size address space */
	start_scan_addr = (unsigned long)ioremap(flash_start, flash_size);
	if (!start_scan_addr) {
		printk ("%s:Failed to ioremap address:0x%x\n",
			__FUNCTION__, flash_start);
		return -EIO;
	}

	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if(mtd_size >= flash_size)
			break;

		printk ("%s: chip probing count %d\n",
			__FUNCTION__, idx);

		map_banks[idx] = (struct map_info *)kmalloc(sizeof(struct map_info), GFP_KERNEL);
		if(map_banks[idx] == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);
		if(map_banks[idx]->name == NULL) {
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);

		sprintf(map_banks[idx]->name, "INCA-IP Bank %d", idx);
		map_banks[idx]->buswidth = 2;
		map_banks[idx]->read8 = incaip_read8;
		map_banks[idx]->read16 = incaip_read16;
		map_banks[idx]->read32 = incaip_read32;
		map_banks[idx]->copy_from = incaip_copy_from;
		map_banks[idx]->write8 = incaip_write8;
		map_banks[idx]->write16 = incaip_write16;
		map_banks[idx]->write32 = incaip_write32;
		map_banks[idx]->copy_to = incaip_copy_to;
		map_banks[idx]->map_priv_1 =
		start_scan_addr + ((idx > 0) ?
		(mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);

		/* start to probe flash chips */
		probing = 1;
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);
		probing = 0;
		if(mtd_banks[idx]) {
			struct cfi_private *cfi;

			mtd_banks[idx]->module = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;
			cfi = (struct cfi_private *)map_banks[idx]->fldrv_priv;
			/* Swap the address for correct access via EBU */
			cfi->addr_unlock1 ^= 2;
			cfi->addr_unlock2 ^= 2;
			printk("%s: bank%ld, name:%s, size:%dbytes \n", __FUNCTION__, num_banks,
			mtd_banks[idx]->name, mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if(!num_banks)
	{
		printk("INCA-IP: No support flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select Static partition definitions
	 */
	part_banks[0].mtd_part = incaip_partitions;
	part_banks[0].type = "static image";
	part_banks[0].nums = NB_OF(incaip_partitions);
	part_banks[1].mtd_part = incaip_fs_partitions;
	part_banks[1].type = "static file system";
	part_banks[1].nums = NB_OF(incaip_fs_partitions);

	for(idx = 0; idx < num_banks ; idx++) {
#ifdef CONFIG_MTD_CMDLINE_PARTS
		sprintf(mtdid, "%d", idx);
		n = parse_cmdline_partitions(mtd_banks[idx],
					     &part_banks[idx].mtd_part,
					     mtdid);
		debugk ("%s: %d command line partitions on bank %s\n",
			__FUNCTION__, n, mtdid);
		if (n > 0) {
			part_banks[idx].type = "command line";
			part_banks[idx].nums = n;
		}
#endif
		if (part_banks[idx].nums == 0) {
			printk (KERN_NOTICE "INCA-IP flash%d: "
				"no partition info available, "
				"registering whole flash at once\n", idx);
			add_mtd_device(mtd_banks[idx]);
		} else {
			printk (KERN_NOTICE "INCA-IP flash%d: "
				"Using %s partition definition\n",
				idx, part_banks[idx].type);
			add_mtd_partitions(mtd_banks[idx],
					   part_banks[idx].mtd_part,
					   part_banks[idx].nums);
		}
	}
#else
	printk (KERN_NOTICE "INCA-IP flash: registering %d whole "
		"flash banks at once\n", num_banks);
	for(idx = 0 ; idx < num_banks ; idx++)
		add_mtd_device(mtd_banks[idx]);
#endif
	return 0;
error_mem:
	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if(map_banks[idx] != NULL) {
			if(map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}
error:
	iounmap((void *)start_scan_addr);
	return ret;
}

static void __exit cleanup_incaip_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++)
	{
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
			del_mtd_partitions(mtd_banks[idx]);
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}
	if (start_scan_addr) {
		iounmap((void *)start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_incaip_mtd);
module_exit(cleanup_incaip_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("MTD map driver for INCA-IP boards");
