/*
 * Driver for the INCAIP Switch
 * Copyright (C) 2003 Gary Jennejohn >gj@denx.de>
 *
 * Based on:
 *
 * isa-skeleton.c: A network driver outline for linux.
 *
 *	Written 1993-94 by Donald Becker.
 *
 *	Copyright 1993 United States Government as represented by the
 *	Director, National Security Agency.
 *
 *	This software may be used and distributed according to the terms
 *	of the GNU General Public License, incorporated herein by reference.
 *
 *	The author may be reached as becker@scyld.com, or C/O
 *	Scyld Computing Corporation
 *	410 Severn Ave., Suite 210
 *	Annapolis MD 21403
 *
 */

static const char *version = "ncaipsw Ver. 0.1\n";

/*
 *  Sources:
 *	List your sources of programming information to document that
 *	the driver is your own creation, and give due credit to others
 *	that contributed to the work. Remember that GNU project code
 *	cannot use proprietary or trade secret information. Interface
 *	definitions are generally considered non-copyrightable to the
 *	extent that the same names and structures must be used to be
 *	compatible.
 *
 *	Finally, keep in mind that the Linux kernel is has an API, not
 *	ABI. Proprietary object-code-only distributions are not permitted
 *	under the GPL.
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <linux/init.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <asm/incaip/dma.h>
/* TEST */
#include <net/tcp.h>

/*
 * The name of the card. Is used for messages and in the requests for
 * io regions, irqs and dma channels
 */
static const char* cardname = "incaipsw";

/* First, a few definitions that the brave might change. */

/* use 0 for production, 1 for verification, >2 for debug */
#ifndef NET_DEBUG
#define NET_DEBUG 2
#endif
static unsigned int incaipsw_debug = NET_DEBUG;

#define MY_TX_TIMEOUT  ((400*HZ)/1000)

/* Information that need to be kept for each board. */
struct incaipsw_local {
	struct net_device_stats stats;
	long open_time;			/* Useless example local info. */

	/* Tx control lock.  This protects the transmit buffer ring
	 * state along with the "tx full" state of the driver.  This
	 * means all netif_queue flow control actions are protected
	 * by this lock as well.
	 */
	spinlock_t lock;
};

#define NUM_RX_DESC	31
#define NUM_TX_DESC	15
#define PKTSIZE_ALIGN	1520
#define INCAIP_SW_DMACHAN 0

#define SW_WRITE_REG(reg, value)   \
         *((volatile u32*)reg) = (u32)value;\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         *((volatile u32*)reg) = (u32)value;

#define SW_READ_REG(reg, value)    \
         value = (u32)*((volatile u32*)reg);\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         value = (u32)*((volatile u32*)reg);

/* Index to functions, as function prototypes. */

extern int	incaipsw_probe(struct net_device *dev);

static int	incaipsw_open(struct net_device *dev);
static int	incaipsw_send_packet(struct sk_buff *skb, struct net_device *dev);
static void	incaipsw_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static void	incaipsw_rx(struct net_device *dev);
static int	incaipsw_close(struct net_device *dev);
static struct	net_device_stats *incaipsw_get_stats(struct net_device *dev);
#if 0
static void	incaipsw_set_multicast_list(struct net_device *dev);
#endif
static void     incaipsw_tx_timeout(struct net_device *dev);
static unsigned long incaip_sw_allocbuf(void *);
static unsigned long incaip_sw_rcv(void *, void *, unsigned long);
static unsigned long incaip_sw_txc(void *);


/* Example routines you must write ;->. */
#define tx_done(dev) 1
static void	hardware_send_packet(struct net_device *dev, char *buf, int length);
static void 	chipset_init(struct net_device *dev, int startp);

static void incaipsw_dma_init(void)
{
	incaDMA_Init();
}

static int tx_full(struct net_device *dev)
{
	return(0);
}

static void hardware_send_packet(struct net_device *dev, char *buf, int length)
{
	/* should check the return value */
	incaDMA_Transmit(INCAIP_SW_DMACHAN, buf, length, 1);
	return;
}

static void chipset_init(struct net_device *dev, int startp)
{
/* resetting and reconfiguring doesn't work */
#if 0
	u32 regValue;

	/* reset the MDIO */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_CFG, (1 << 31));
	/* wait for reset to finish */
	do
	{
		SW_READ_REG(INCA_IP_Switch_MDIO_CFG, regValue);
	} while (regValue & (1 << 31));

	/* To workaround a problem with collision counter
	 * (see Errata sheet).
	 */
	SW_WRITE_REG(INCA_IP_Switch_PC_TX_CTL, 0x00000001);
	SW_WRITE_REG(INCA_IP_Switch_LAN_TX_CTL, 0x00000001);

	   /* init MDIO configuration: 
	         MDS (Poll speed):       0x01 (4ms)
	         PHY_LAN_ADDR:           0x06
	         PHY_PC_ADDR:            0x05
	         UEP (Use External PHY): 0x00 (Internal PHY is used)
	         PS (Port Select):       0x00 (PT/UMM for LAN)
	         PT (PHY Test):          0x00 (no test mode)
	         UMM (Use MDIO Mode):    0x00 (state machine is disabled)
	   */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_CFG, 0x4c50);

	   /* init PHY: 
	         SL (Auto Neg. Speed for LAN)
	         SP (Auto Neg. Speed for PC)
	         LL (Link Status for LAN)
	         LP (Link Status for PC)
	         DL (Duplex Status for LAN)
	         DP (Duplex Status for PC)
	         PL (Auto Neg. Pause Status for LAN)
	         PP (Auto Neg. Pause Status for PC)
	   */
	SW_WRITE_REG (INCA_IP_Switch_EPHY, 0xff);

	   /* MDIO_ACC:
	         RA (Request/Ack)  0x01 (Request)
	         RW (Read/Write)   0x01 (Write)
	         PHY_ADDR          0x05 (PC)
	         REG_ADDR          0x00 (PHY_BCR: basic control register)
	         PHY_DATA          0x8000
	                              Reset                   - software reset
	                              LB (loop back)          - normal
	                              SS (speed select)       - 10 Mbit/s
	                              ANE (auto neg. enable)  - disable
	                              PD (power down)         - normal
	                              ISO (isolate)           - normal
	                              RAN (restart auto neg.) - normal
	                              DM (duplex mode)        - half duplex
	                              CT (collision test)     - enable
	   */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC, 0xc0a08000);

	   /* MDIO_ACC:
	         RA (Request/Ack)  0x01 (Request)
	         RW (Read/Write)   0x01 (Write)
	         PHY_ADDR          0x06 (LAN)
	         REG_ADDR          0x00 (PHY_BCR: basic control register)
	         PHY_DATA          0x8000
	                              Reset                   - software reset
	                              LB (loop back)          - normal
	                              SS (speed select)       - 10 Mbit/s
	                              ANE (auto neg. enable)  - disable
	                              PD (power down)         - normal
	                              ISO (isolate)           - normal
	                              RAN (restart auto neg.) - normal
	                              DM (duplex mode)        - half duplex
	                              CT (collision test)     - enable
	   */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC, 0xc0c08000);

	/* Make sure the CPU port is disabled for now. We
	 * don't want packets to get stacked for us until
	 * we enable DMA and are prepared to receive them.
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf);

	SW_READ_REG(INCA_IP_Switch_ARL_CTL, regValue);

	/* CRC GEN is enabled.
	 */
	regValue |= 0x00000200;
	SW_WRITE_REG(INCA_IP_Switch_ARL_CTL, regValue);

	/* ADD TAG is disabled.
	 */
	SW_READ_REG(INCA_IP_Switch_PMAC_HD_CTL, regValue);
	regValue &= ~0x00000002;
	SW_WRITE_REG(INCA_IP_Switch_PMAC_HD_CTL, regValue);
#endif
}

static unsigned char my_ethaddr[MAX_ADDR_LEN];
/* need to get the ether addr from u-boot */
static int __init ethaddr_setup(char *line)
{
	char *ep;
	int i;

	memset(my_ethaddr, 0, MAX_ADDR_LEN);
	/* there should really be routines to do this stuff */
	for (i = 0; i < 6; i++)
	{
		my_ethaddr[i] = line ? simple_strtoul(line, &ep, 16) : 0;
		if (line)
			line = (*ep) ? ep+1 : ep;
	}
	return 0;
}
__setup("ethaddr=", ethaddr_setup);

/*
 * The INCAIP Switch is always there - it's in the CPU.
 */
int __init 
incaipsw_probe(struct net_device *dev)
{
	struct incaipsw_local *np;
	static unsigned version_printed;
	int i, retval;
	unsigned short wTmp;
	unsigned long regValue;
	static int init_done = 0;
	int irqval;

	/* otherwise we return success for all possible interfaces! */
	if (init_done != 0)
		return 1;

	SET_MODULE_OWNER(dev);

	if (incaipsw_debug  &&  version_printed++ == 0)
		printk(KERN_DEBUG "%s", version);

	printk(KERN_INFO "%s: %s, ", dev->name, cardname);

	/* this should be handled by ethaddr_setup() */
	/* TEST */
	retval = 0;
	for (i = 0; i < 6; i++)
		retval += my_ethaddr[i];
	/* ethaddr not set in u-boot ? */
	if (retval == 0)
	{
		dev->dev_addr[0] = 0x00;
		dev->dev_addr[1] = 0x20;
		dev->dev_addr[2] = 0xda;
		dev->dev_addr[3] = 0x86;
		dev->dev_addr[4] = 0x23;
		dev->dev_addr[5] = 0x74;
	}
	else
		for (i = 0; i < 6; i++)
			dev->dev_addr[i] = my_ethaddr[i];
	/* print it out */
	for (i = 0; i < 6; i++)
		printk(" %2.2x", dev->dev_addr[i]);
	printk("\n");

	dev->irq = INT_NUM_IM1_IRL7;

	incaipsw_dma_init();

	/* let the DMA driver allocate space for us */
	incaDMA_InitRxChannel(INCAIP_SW_DMACHAN, NUM_RX_DESC, PKTSIZE_ALIGN, dev,
		incaip_sw_allocbuf, incaip_sw_rcv);
	incaDMA_InitTxChannel(INCAIP_SW_DMACHAN, NUM_TX_DESC, 3, dev, incaip_sw_txc);

	/* Set MAC address.
	 */
	wTmp = (u16)dev->dev_addr[0];
	regValue = (wTmp << 8) | dev->dev_addr[1];

	SW_WRITE_REG(INCA_IP_Switch_PMAC_SA1, regValue);

	wTmp = (u16)dev->dev_addr[2];
	regValue = (wTmp << 8) | dev->dev_addr[3];
	regValue = regValue << 16;
	wTmp = (u16)dev->dev_addr[4];
	regValue |= (wTmp<<8) | dev->dev_addr[5];

	SW_WRITE_REG(INCA_IP_Switch_PMAC_SA2, regValue);

	/* enable spanning tree forwarding, enable the CPU port */
	/* ST_PT:
	      CPS (CPU port status)   0x3 (forwarding)
	      LPS (LAN port status)   0x3 (forwarding)
	      PPS (PC port status)    0x3 (forwarding)
	*/
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0x3f);


	/* Initialize the device structure. */
	if (dev->priv == NULL) {
		dev->priv = kmalloc(sizeof(struct incaipsw_local), GFP_KERNEL);
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	memset(dev->priv, 0, sizeof(struct incaipsw_local));

	np = (struct incaipsw_local *)dev->priv;
	spin_lock_init(&np->lock);

#if 0 /* needed? */
	/* Grab the region so that no one else tries to probe our ioports. */
	request_region(ioaddr, NETCARD_IO_EXTENT, cardname);
#endif

	dev->open		= incaipsw_open;
	dev->stop		= incaipsw_close;
	dev->hard_start_xmit	= incaipsw_send_packet;
	dev->get_stats		= incaipsw_get_stats;
#if 0
	dev->set_multicast_list = incaipsw_set_multicast_list;
#endif

        dev->tx_timeout		= incaipsw_tx_timeout;
        dev->watchdog_timeo	= MY_TX_TIMEOUT; 

	/* Fill in the fields of the device structure with ethernet values. */
	ether_setup(dev);

	init_done = 1;

	return 0;
}

static void incaipsw_tx_timeout(struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;

	printk(KERN_WARNING "%s: transmit timed out, %s?\n", dev->name,
	       tx_done(dev) ? "IRQ conflict" : "network cable problem");

	/* Try to restart the adaptor. */
	chipset_init(dev, 1);

	np->stats.tx_errors++;

	/* If we have space available to accept new transmit
	 * requests, wake up the queueing layer.  This would
	 * be the case if the chipset_init() call above just
	 * flushes out the tx queue and empties it.
	 *
	 * If instead, the tx queue is retained then the
	 * netif_wake_queue() call should be placed in the
	 * TX completion interrupt handler of the driver instead
	 * of here.
	 */
	if (!tx_full(dev))
		netif_wake_queue(dev);
}

/*
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
static int
incaipsw_open(struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;
#ifdef INCAIPSW_DEBUG
printk("%s called\n", __FUNCTION__);
#endif

	/*
	 * This is used if the interrupt line can turned off (shared).
	 * See 3c503.c for an example of selecting the IRQ at config-time.
	 */
	if (request_irq(dev->irq, &incaipsw_interrupt, 0, cardname, dev)) {
		return -EAGAIN;
	}

	/* Reset the hardware here. Don't forget to set the station address. */
	chipset_init(dev, 1);
	np->open_time = jiffies;

	/* We are now ready to accept transmit requeusts from
	 * the queueing layer of the networking.
	 */
	netif_start_queue(dev);

	return 0;
}

/* This will only be invoked if your driver is _not_ in XOFF state.
 * What this means is that you need not check it, and that this
 * invariant will hold if you make sure that the netif_*_queue()
 * calls are done at the proper times.
 */
static int incaipsw_send_packet(struct sk_buff *skb, struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;
	short length = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
	unsigned char *buf = (unsigned char *)KSEG1ADDR(skb->data);
	struct tcphdr *th;
#ifdef INCAIPSW_TXDEBUG
	int command;
#endif

#ifdef INCAIPSW_DEBUG
printk("%s called\n", __FUNCTION__);
#endif
	/* If some error occurs while trying to transmit this
	 * packet, you should return '1' from this function.
	 * In such a case you _may not_ do anything to the
	 * SKB, it is still owned by the network queueing
	 * layer when an error is returned.  This means you
	 * may not modify any SKB fields, you may not free
	 * the SKB, etc.
	 */

	/* This is the case for older hardware which takes
	 * a single transmit buffer at a time, and it is
	 * just written to the device via PIO.
	 *
	 * No spin locking is needed since there is no TX complete
	 * event.  If by chance your card does have a TX complete
	 * hardware IRQ then you may need to utilize np->lock here.
	 */
#ifdef INCAIP_SW_TXDEBUG
if (skb->nh.iph->protocol == 6)
{
printk("6 ");
printk("Sack: %#x ", skb->h.th->ack);
printk("Spsh: %#x ", skb->h.th->psh);
printk("Srst: %#x ", skb->h.th->rst);
printk("Ssyn: %#x ", skb->h.th->syn);
printk("Sfin: %#x ", skb->h.th->fin);
printk("\n");
}
#endif
#ifdef INCAIPSW_TXDEBUG
printk("TX: ");
for (command = 0; command < skb->len; command++)
{
printk("0x%02X ", buf[command]);
if(((command+1)%16) == 0)printk("\n");
}
printk("\n");
#endif
#ifndef CONFIG_MIPS_UNCACHED
	dma_cache_wback_inv(skb->data, length);
#endif
	hardware_send_packet(dev, buf, length);
	np->stats.tx_bytes += skb->len;
	np->stats.tx_packets++;

	dev->trans_start = jiffies;

	dev_kfree_skb (skb);

	return 0;
}

/*
 * The typical workload of the driver:
 * Handle the network interface interrupts.
 *
 * With this hardware all the real work is handled by the DMA driver,
 * so there;s nothing to do here.
 *
 */
static void incaipsw_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
#ifdef INCAIPSW_DEBUG
	printk("%s\n", __FUNCTION__);
#endif
	return;
}

/* The inverse routine to incaipsw_open(). */
static int
incaipsw_close(struct net_device *dev)
{
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;
	int ioaddr = dev->base_addr;

	lp->open_time = 0;

	netif_stop_queue(dev);

	/* Flush the Tx and disable Rx here. */

	/* Disable forwarding to the CPU port.
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf);

	/* Close RxDMA channel.
	 */
	incaDMA_CloseRxChannel(0);
	incaDMA_CloseTxChannel(0);

	free_irq(dev->irq, dev);

	/* Update the statistics here. */

	return 0;

}

/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
static struct net_device_stats *incaipsw_get_stats(struct net_device *dev)
{
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;

	return &lp->stats;
}

/*
 * Set or clear the multicast filter for this adaptor.
 * num_addrs == -1	Promiscuous mode, receive all packets
 * num_addrs == 0	Normal mode, clear multicast list
 * num_addrs > 0	Multicast mode, receive normal and MC packets,
 *			and do best-effort filtering.
 */
#if 0
static void
incaipsw_set_multicast_list(struct net_device *dev)
{
	short ioaddr = dev->base_addr;
	if (dev->flags&IFF_PROMISC)
	{
		/* Enable promiscuous mode */
		outw(MULTICAST|PROMISC, ioaddr);
	}
	else if((dev->flags&IFF_ALLMULTI) || dev->mc_count > HW_MAX_ADDRS)
	{
		/* Disable promiscuous mode, use normal mode. */
		hardware_set_filter(NULL);

		outw(MULTICAST, ioaddr);
	}
	else if(dev->mc_count)
	{
		/* Walk the address list, and load the filter */
		hardware_set_filter(dev->mc_list);

		outw(MULTICAST, ioaddr);
	}
	else 
		outw(0, ioaddr);
}
#endif

static unsigned long incaip_sw_allocbuf(void *dev)
{
#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif
	return 0;
}

static unsigned long incaip_sw_rcv(void *vdev, void *buf, unsigned long size)
{
	/* Malloc up new buffer. */
	struct sk_buff *skb;
	struct net_device *dev = (struct net_device *)vdev;
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;
#ifndef CONFIG_MIPS_UNCACHED
	unsigned char *lbuf = (unsigned char *)buf;
#else
	unsigned char *lbuf = (unsigned char *)KSEG1ADDR(buf);
#endif
	unsigned char *ldat;
#ifdef INCASW_RX_DEBUG
	int command;
#endif

#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif

	lp->stats.rx_bytes+=size;
	
	skb = dev_alloc_skb(size + 2);	/* +2 to allow for alignment */
	if (skb == NULL) {
		printk(KERN_NOTICE "%s: Memory squeeze, dropping packet.\n",
			   dev->name);
		lp->stats.rx_dropped++;
		return -ENOMEM;
	}

	/* Make sure IP header is aligned on 16 byte boundary */
	skb_reserve(skb, 2);

	skb->dev = dev;

	skb_put(skb, size);
	ldat = skb->data;
	/* 'skb->data' points to the start of sk_buff data area. */
	memcpy(ldat, lbuf, size);

#ifdef INCASW_RX_DEBUG
printk("RX: ");
for (command = 0; command < size; command++)
{
printk("0x%02X ", lbuf[command]);
if(((command+1)%16) == 0)printk("\n");
}
printk("\n");
#endif

        skb->protocol=eth_type_trans(skb,dev);
	skb->pkt_type = PACKET_HOST;
#ifdef INCAIP_SW_RXDEBUG
if (ldat[14] == 0x45 && ldat[23] == 6)
{
printk("6 ");
printk("Rack: %#x ", ((struct tcphdr *)&ldat[34])->ack);
printk("Rpsh: %#x ", ((struct tcphdr *)&ldat[34])->psh);
printk("Rrst: %#x ", ((struct tcphdr *)&ldat[34])->rst);
printk("Rsyn: %#x ", ((struct tcphdr *)&ldat[34])->syn);
printk("Rfin: %#x ", ((struct tcphdr *)&ldat[34])->fin);
printk("\n");
for (command = 0; command < size; command++)
{
printk("0x%02X ", ldat[command]);
if(((command+1)%16) == 0)printk("\n");
}
printk("\n");
}
#endif
	netif_rx(skb);
	dev->last_rx = jiffies;
	lp->stats.rx_packets++;
	return 0;
}

static unsigned long incaip_sw_txc(void *vdev)
{
	struct net_device *dev = (struct net_device *)vdev;

#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif
	return 0;
}


#ifdef MODULE

static struct net_device this_device;
MODULE_LICENSE("GPL");

int init_module(void)
{
	int result;

	/* the INCAIP switch is always there, so really no need to probe */
	this_device.init      = netcard_probe;

	if ((result = register_netdev(&this_device)) != 0)
		return result;

	return 0;
}

void
cleanup_module(void)
{
	/* Disable forwarding to the CPU port.
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf);

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	unregister_netdev(&this_device);

	/*
	 * If we don't do this, we can't re-insmod it later.
	 * Release irq here, when you have jumpered versions and
	 * allocate them in incaipsw_probe().
	 */
	free_irq(this_device.irq, dev);

	/* Close DMA channels.
	 */
	incaDMA_CloseRxChannel(0);
	incaDMA_CloseTxChannel(0);

	if (this_device.priv)
		kfree(this_device.priv);
}

#endif /* MODULE */
